/*Copyright 2003 The Board of Trustees of The Leland Stanford Junior University.
  All Rights Reserved.

  Communication to Jessica Shapiro (jessicas@stanford.edu)
*/


#ifdef DISPLAYON

#include <stdio.h>
#include <stdlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <tk.h>
#include <tcl.h>
#include "display.h"


Display *mydisplay;
Window GLwin;
GLXContext cx;
Tk_Window GLwin_tk;


struct _WINDOWINFO {
  int x, y;
  int width, height;
} windInfo = {
  -1, -1, 100, 100
};


static struct {
  int keyField;
  void (*KeyFunc)(void);
} keyTable[200];


static struct {
  int mouseField;
  void (*MouseFunc)(AUX_EVENTREC *);
} mouseDownTable[20], mouseUpTable[20], mouseLocTable[20];


extern Tcl_Interp *interp;
extern char* lockdir;


static int keyTableCount = 0;
static int mouseDownTableCount = 0;
static int mouseUpTableCount = 0;
static int mouseLocTableCount = 0;

static int lastEventType = -1;
static GLenum MouseLoc(int, int, GLenum);
static GLenum MouseUp(int, int, GLenum);
static GLenum MouseDown(int, int, GLenum);
static GLenum KeyDown(int, GLenum);
static void (*ExposeFunc)(int, int) = 0;
static void (*ReshapeFunc)(int, int) = 0;
static void (*DisplayFunc)(void) = 0;
static GLenum (*KeyDownFunc)(int, GLenum) = KeyDown;
static GLenum (*MouseDownFunc)(int, int, GLenum) = MouseDown;
static GLenum (*MouseUpFunc)(int, int, GLenum) = MouseUp;
static GLenum (*MouseMoveFunc)(int, int, GLenum) = MouseLoc;
static void (*IdleFunc)(void) = 0;

static int WaitForMapNotify(Display*, XEvent*, char*);
void setupGLWindow(char *);
void addKeyFunc(int key, void (*Func)(void));
void addMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *));
void doGraphicsEvent(ClientData clientdata, XEvent *current);
void InitPosition(int, int, int, int);
void addReshapeFunc(void (*Func)(GLsizei, GLsizei));


static int WaitForMapNotify(Display *d, XEvent *e, char *arg)
{

    if (e->type == MapNotify && e->xmap.window == GLwin) {
	return GL_TRUE;
    }
    return GL_FALSE;
}



void setupGLWindow(char *title) {

  XSetWindowAttributes swa;
  XVisualInfo *vi;
  Colormap cmap;
  unsigned long mask;
  /*  int attributeList[] = {GLX_RGBA, None}; */
  int attributeList[] = {GLX_DOUBLEBUFFER, GLX_RGBA, GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, GLX_DEPTH_SIZE, 1, None};
  XSizeHints sh;
  XEvent evt;


  GLwin_tk = (Tk_Window)(TkCreateMainWindow(interp, NULL, title));
    
  mydisplay = Tk_Display(GLwin_tk);
  
  vi = glXChooseVisual(mydisplay, DefaultScreen(mydisplay), attributeList);
  cx = glXCreateContext(mydisplay, vi, 0, GL_TRUE);    
  cmap = XCreateColormap(mydisplay, RootWindow(mydisplay, vi->screen), vi->visual, AllocNone);
    
  Tk_SetWindowVisual(GLwin_tk, vi->visual, vi->depth, cmap);
  Tk_MakeWindowExist(GLwin_tk);
  GLwin = Tk_WindowId(GLwin_tk);
  
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask |
    ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
  mask = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap;
  XChangeWindowAttributes(mydisplay, GLwin, mask, &swa);
  
  XMoveResizeWindow(mydisplay, GLwin, windInfo.x, windInfo.y, windInfo.width, windInfo.height);
  
  if (windInfo.x != -1 && windInfo.y != -1) {
    sh.flags = USPosition;
    sh.x = windInfo.x + 10;
    sh.y = windInfo.y + 10;
    
    XSetStandardProperties(mydisplay, GLwin, title, title, None, 0, 0, &sh);
  } else {
    XSetStandardProperties(mydisplay, GLwin, title, title, None, 0, 0, 0);
  }
  
  
  XSelectInput(mydisplay, GLwin, StructureNotifyMask | ExposureMask | KeyPressMask |ButtonPressMask | ButtonReleaseMask | PointerMotionMask); 
  
  XMapWindow(mydisplay, GLwin);
  XIfEvent(mydisplay, &evt, WaitForMapNotify, 0);
  
  glXMakeCurrent(mydisplay, GLwin, cx);


  Tk_CreateEventHandler(GLwin_tk, swa.event_mask, doGraphicsEvent, (ClientData)NULL);
  
}  




void addKeyFunc(int key, void (*Func)(void))
{
    keyTable[keyTableCount].keyField = key;
    keyTable[keyTableCount++].KeyFunc = Func;
}


void addMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *))
{
    if (mode == AUX_MOUSEDOWN) {
	mouseDownTable[mouseDownTableCount].mouseField = mouse;
	mouseDownTable[mouseDownTableCount++].MouseFunc = Func;
    } else if (mode == AUX_MOUSEUP) {
	mouseUpTable[mouseUpTableCount].mouseField = mouse;
	mouseUpTable[mouseUpTableCount++].MouseFunc = Func;
    } else if (mode == AUX_MOUSELOC) {
	mouseLocTable[mouseLocTableCount].mouseField = mouse;
	mouseLocTable[mouseLocTableCount++].MouseFunc = Func;
    } 
}



static GLenum MouseLoc(int x, int y, GLenum button) {
  AUX_EVENTREC info;
  GLenum flag;
  int i;
  
  flag = GL_FALSE;
  for (i = 0; i < mouseLocTableCount; i++) {
    if ((button & AUX_LEFTBUTTON) == mouseLocTable[i].mouseField) {
      info.event = AUX_MOUSELOC;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
      (*mouseLocTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_RIGHTBUTTON) == mouseLocTable[i].mouseField) {
      info.event = AUX_MOUSELOC;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
      (*mouseLocTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_MIDDLEBUTTON) == mouseLocTable[i].mouseField) {
      info.event = AUX_MOUSELOC;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
      (*mouseLocTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
  }
  return flag;
}

static GLenum MouseUp(int x, int y, GLenum button) {
  AUX_EVENTREC info;
  GLenum flag;
  int i;
  
  flag = GL_FALSE;
  for (i = 0; i < mouseUpTableCount; i++) {
    if ((button & AUX_LEFTBUTTON) == mouseUpTable[i].mouseField) {
      info.event = AUX_MOUSEUP;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
      (*mouseUpTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_RIGHTBUTTON) == mouseUpTable[i].mouseField) {
      info.event = AUX_MOUSEUP;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
      (*mouseUpTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_MIDDLEBUTTON) == mouseUpTable[i].mouseField) {
      info.event = AUX_MOUSEUP;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
      (*mouseUpTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
  }
  return flag;
}

static GLenum MouseDown(int x, int y, GLenum button) {
  AUX_EVENTREC info;
  GLenum flag;
  int i;
  
  flag = GL_FALSE;
  for (i = 0; i < mouseDownTableCount; i++) {
    if ((button & AUX_LEFTBUTTON) == mouseDownTable[i].mouseField) {
      info.event = AUX_MOUSEDOWN;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
      (*mouseDownTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_RIGHTBUTTON) == mouseDownTable[i].mouseField) {
      info.event = AUX_MOUSEDOWN;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
      (*mouseDownTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
    if ((button & AUX_MIDDLEBUTTON) == mouseDownTable[i].mouseField) {
      info.event = AUX_MOUSEDOWN;
      info.data[AUX_MOUSEX] = x;
      info.data[AUX_MOUSEY] = y;
      info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
      (*mouseDownTable[i].MouseFunc)(&info);
      flag |= GL_TRUE;
    }
  }
  return flag;
}

static GLenum KeyDown(int key, GLenum status) {
  GLenum flag;
  int i;
  
  flag = GL_FALSE;
  printf("inKeyDown\n");
  if (keyTableCount) {
    for (i = 0; i < keyTableCount; i++) {
      if (key == keyTable[i].keyField) {
	(*keyTable[i].KeyFunc)();
	flag |= GL_TRUE;
      }
    }
  }
  return flag;
}


void doGraphicsEvent(ClientData clientdata, XEvent *current) {
  XEvent ahead;
  char buf[1000];
  KeySym ks;
  int key;
  
  fprintf(stderr, "dographicsevent\n");fflush(stderr);
  switch (current->type) {
  case MappingNotify:
    XRefreshKeyboardMapping((XMappingEvent *)current);
    lastEventType = MappingNotify;
    return;
    
    
  case Expose:
    while (XEventsQueued(current->xexpose.display, QueuedAfterReading) > 0) {
      XPeekEvent(current->xexpose.display, &ahead);
      if (ahead.xexpose.window != current->xexpose.window ||
	  ahead.type != Expose) {
	break;
      }
      XNextEvent(mydisplay, current);
    }
    if (current->xexpose.count == 0) {
      if (ExposeFunc) {
	(*ExposeFunc)(windInfo.width, windInfo.height);
	if (lastEventType == ConfigureNotify) {
	  lastEventType = Expose;
	  return;
	} else {
	  lastEventType = Expose;
	  return;
	}
      }
    }
    return;
    
  case ConfigureNotify:
    lastEventType = ConfigureNotify;
    windInfo.width = current->xconfigure.width;
    windInfo.height = current->xconfigure.height;
    if (ReshapeFunc) {
      (*ReshapeFunc)(windInfo.width, windInfo.height);
      return;
    } else {
      return;
    }
    
  case MotionNotify:
    lastEventType = MotionNotify;
    if (MouseMoveFunc) {
      GLenum mask;
      
      mask = 0;
      if (current->xmotion.state & Button1Mask) {
	mask |= AUX_LEFTBUTTON;
      }
      if (current->xmotion.state & Button2Mask) {
	mask |= AUX_MIDDLEBUTTON;
      }
      if (current->xmotion.state & Button3Mask) {
	mask |= AUX_RIGHTBUTTON;
      }
      (*MouseMoveFunc)(current->xmotion.x, current->xmotion.y, mask);
      return;
    } else {
      return;
    }
    
  case ButtonPress:

    lastEventType = ButtonPress;
    if (MouseDownFunc) {
      GLenum mask;
      
      mask = 0;
      if (current->xbutton.button == 1) {
	mask |= AUX_LEFTBUTTON;
      }
      if (current->xbutton.button == 2) {
	mask |= AUX_MIDDLEBUTTON;
      }
      if (current->xbutton.button == 3) {
	mask |= AUX_RIGHTBUTTON;
      }
      (*MouseDownFunc)(current->xbutton.x, current->xbutton.y, mask);
      return;
    } else {
      return;
    }
  case ButtonRelease:
    lastEventType = ButtonRelease;
    
    if (MouseUpFunc) {
      GLenum mask;
      
      mask = 0;
      if (current->xbutton.button == 1) {
	mask |= AUX_LEFTBUTTON;
      }
      if (current->xbutton.button == 2) {
	mask |= AUX_MIDDLEBUTTON;
      }
      if (current->xbutton.button == 3) {
	mask |= AUX_RIGHTBUTTON;
      }
      (*MouseUpFunc)(current->xbutton.x, current->xbutton.y, mask);
      return;
    } else {
      return;
    }
    
  case KeyPress:

    lastEventType = KeyPress;
    XLookupString(&(current->xkey), buf, sizeof(buf), &ks, 0);
    
    switch (ks) {
    case AUX_0: 		key = AUX_0;		break;
    case AUX_1: 		key = AUX_1;		break;
    case AUX_2: 		key = AUX_2;		break;
    case AUX_3: 		key = AUX_3;		break;
    case AUX_4: 		key = AUX_4;		break;
    case AUX_5: 		key = AUX_5;		break;
    case AUX_6: 		key = AUX_6;		break;
    case AUX_7: 		key = AUX_7;		break;
    case AUX_8: 		key = AUX_8;		break;
    case AUX_9: 		key = AUX_9;		break;
    case AUX_A: 		key = AUX_A;		break;
    case AUX_B: 		key = AUX_B;		break;
    case AUX_C: 		key = AUX_C;		break;
    case AUX_D: 		key = AUX_D;		break;
    case AUX_E: 		key = AUX_E;		break;
    case AUX_F: 		key = AUX_F;		break;
    case AUX_G: 		key = AUX_G;		break;
    case AUX_H: 		key = AUX_H;		break;
    case AUX_I: 		key = AUX_I;		break;
    case AUX_J: 		key = AUX_J;		break;
    case AUX_K: 		key = AUX_K;		break;
    case AUX_L: 		key = AUX_L;		break;
    case AUX_M: 		key = AUX_M;		break;
    case AUX_N: 		key = AUX_N;		break;
    case AUX_O: 		key = AUX_O;		break;
    case AUX_P: 		key = AUX_P;		break;
    case AUX_Q: 		key = AUX_Q;		break;
    case AUX_R: 		key = AUX_R;		break;
    case AUX_S: 		key = AUX_S;		break;
    case AUX_T: 		key = AUX_T;		break;
    case AUX_U: 		key = AUX_U;		break;
    case AUX_V: 		key = AUX_V;		break;
    case AUX_W: 		key = AUX_W;		break;
    case AUX_X: 		key = AUX_X;		break;
    case AUX_Y: 		key = AUX_Y;		break;
    case AUX_Z: 		key = AUX_Z;		break;
    case AUX_a: 		key = AUX_a;		break;
    case AUX_b: 		key = AUX_b;		break;
    case AUX_c: 		key = AUX_c;		break;
    case AUX_d: 		key = AUX_d;		break;
    case AUX_e: 		key = AUX_e;		break;
    case AUX_f: 		key = AUX_f;		break;
    case AUX_g: 		key = AUX_g;		break;
    case AUX_h: 		key = AUX_h;		break;
    case AUX_i: 		key = AUX_i;		break;
    case AUX_j: 		key = AUX_j;		break;
    case AUX_k: 		key = AUX_k;		break;
    case AUX_l: 		key = AUX_l;		break;
    case AUX_m: 		key = AUX_m;		break;
    case AUX_n: 		key = AUX_n; 		break;
    case AUX_o: 		key = AUX_o;		break;
    case AUX_p: 		key = AUX_p;		break;
    case AUX_q: 		key = AUX_q;		break;
    case AUX_r: 		key = AUX_r;		break;
    case AUX_s: 		key = AUX_s;		break;
    case AUX_t: 		key = AUX_t;		break;
    case AUX_u: 		key = AUX_u;		break;
    case AUX_v: 		key = AUX_v;		break;
    case AUX_w: 		key = AUX_w;		break;
    case AUX_x: 		key = AUX_x;		break;
    case AUX_y: 		key = AUX_y;		break;
    case AUX_z: 		key = AUX_z;		break;
    case AUX_SPACE:	        key = AUX_SPACE;	break;
    case AUX_RETURN: 	        key = AUX_RETURN;	break;
    case AUX_ESCAPE: 	        key = AUX_ESCAPE;	break;
    case AUX_LEFT:	        key = AUX_LEFT;		break;
    case AUX_UP:		key = AUX_UP;		break;
    case AUX_RIGHT:  	        key = AUX_RIGHT;	break;
    case AUX_DOWN:	        key = AUX_DOWN;		break;
    default: 		        key = GL_FALSE;		break;
    }
    if (key && KeyDownFunc) {
      GLenum mask;
      
      mask = 0;
      if (current->xkey.state & ControlMask) {
	mask |= AUX_CONTROL;
      }
      if (current->xkey.state & ShiftMask) {
	mask |= AUX_SHIFT;
      }
      (*KeyDownFunc)(key, mask);
      return;
    } else {
      return;
    }
  }
}

void InitPosition(int x, int y, int width, int height) {
  
  windInfo.x = x;
  windInfo.y = y;
  windInfo.width = width;
  windInfo.height = height;

}

void addReshapeFunc(void (*Func)(GLsizei, GLsizei))
{
    ExposeFunc = Func;
    ReshapeFunc = Func;
}

#endif /* DISPLAYON */		   
