/*Copyright 2003 The Board of Trustees of The Leland Stanford Junior University.
  All Rights Reserved.

  Communication to Jessica Ebert (jessicas@stanford.edu)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defines.h"
#include "transform.h"



typedef struct Pair {
  Vector p1, p2;
  float weight;
} Pair;

static Pair **pairs = NULL;
static int npairs = 0;
static int max_pairs;
static int unique_points=0;
static int unique_points_1 = 0; /*added: JCS 04/9/02*/
static int unique_points_2 = 0; /*added: JCS 06/3/02*/

typedef	struct	{
		unsigned int dim, max_dim;
		float *ve;
		} VEC;

typedef	struct	{
		unsigned int m, n;
		unsigned int max_m, max_n, max_size;
		float **me,*base;	/* base is base of alloc'd mem */
		} MAT;

VEC *symmeig(MAT*, MAT*, VEC*);

extern	VEC *v_get(int);
extern	MAT *m_get(int,int);
extern  int m_free(MAT*);
extern int v_free(VEC*);

void convert_quat_to_mat(Quaternion, Matrix);


/*okayToMatchPoints: Added this function on 04/09/02 in order to query the
 *status of the variable unique_points, which makes sure that at least two of
 *the points entered have a different p1 value. Otherwise, match_pairs will
 *hang. It is up to the user to call match_pairs only when appropriate; it
 *does not check unique_points itself.*/
int okayToMatchPoints() {
  if(unique_points >= 2) {
    return 1;
  }else {
    return 0;
  }
}

void vapply (Matrix m, Vector a, Vector b) {
  int j;
  Vector t;

  for (j = 0; j <= 2; j++)
    t [j] = a [0] * m [0] [j] + a [1] * m [1] [j] +
            a [2] * m [2] [j] + m [3] [j];

  for (j = 0; j <= 2; j++)
    b [j] = t [j];
}


void init_pairs() {
  int i;

  if (pairs != NULL) {
    for (i = 0; i < npairs; i++)
      free (pairs[i]);
    free (pairs);
  }

  npairs = 0;
  unique_points = 0;
  max_pairs = 5000;
  pairs = (Pair **) malloc (sizeof (Pair *) * max_pairs);
}



void add_pair(Vector p1, Vector p2, float weight) {
  Pair *mynew;
  int i,distinct = 1;
  float epsilon = 0.000001;

  /*  printf("pair: %.3f %.3f %.3f, %.3f %.3f %.3f\n",p1[X],p1[Y],p1[Z],p2[X],p2[Y],p2[Z]);*/
  if(unique_points<2) {
    for(i=0;i<npairs;i++) {
      if((fabs(pairs[i]->p1[X]-p1[X]) < epsilon) && fabs(pairs[i]->p1[Y]-p1[Y])<epsilon && fabs(pairs[i]->p1[Z]-p1[Z])<epsilon) {
	distinct = 0;
	break;
      }
      if((fabs(pairs[i]->p2[X]-p2[X]) < epsilon) && fabs(pairs[i]->p2[Y]-p2[Y])<epsilon && fabs(pairs[i]->p2[Z]-p2[Z])<epsilon) {      
	distinct=0;
	break;
      }
    }
    if(distinct) {
      unique_points++;
    }
  }


  if (npairs >= max_pairs) {
    max_pairs *= 2;
    pairs = (Pair **) realloc (pairs, sizeof (Pair *) * max_pairs);
  }

  mynew = (Pair *) malloc (sizeof (Pair));
  pairs[npairs] = mynew;
  npairs++;

  mynew->p1[X] = p1[X];
  mynew->p1[Y] = p1[Y];
  mynew->p1[Z] = p1[Z];

  mynew->p2[X] = p2[X];
  mynew->p2[Y] = p2[Y];
  mynew->p2[Z] = p2[Z];

  mynew->weight = weight;
}


void match_pairs(Matrix rotmat, Vector trans)

{
  int i,j,k;
  int dummy;
  Vector c1,c2;	
  Vector v1,v2;
  float recip;
  float tr;
  MAT *m, *q;
  VEC *v;
  Matrix3 cov;
  Matrix3 aij;
  Quaternion quat;
  float weight;
  float weight_sum;




  weight_sum = 0;
  c1[X] = c1[Y] = c1[Z] = 0;
  c2[X] = c2[Y] = c2[Z] = 0;

  for (i = 0; i < npairs; i++) {

    weight = pairs[i]->weight;
    weight_sum += weight;

    c1[X] += pairs[i]->p1[X] * weight;
    c1[Y] += pairs[i]->p1[Y] * weight;
    c1[Z] += pairs[i]->p1[Z] * weight;

    c2[X] += pairs[i]->p2[X] * weight;
    c2[Y] += pairs[i]->p2[Y] * weight;
    c2[Z] += pairs[i]->p2[Z] * weight;
  }

  recip = 1.0 / weight_sum;

  c1[X] *= recip;
  c1[Y] *= recip;
  c1[Z] *= recip;

  c2[X] *= recip;
  c2[Y] *= recip;
  c2[Z] *= recip;

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      cov[i][j] = 0;
    }
  }

  for (k = 0; k < npairs; k++) {

    weight = pairs[k]->weight;

    v1[X] = (pairs[k]->p1[X] - c1[X]) * weight;
    v1[Y] = (pairs[k]->p1[Y] - c1[Y]) * weight;
    v1[Z] = (pairs[k]->p1[Z] - c1[Z]) * weight;

    v2[X] = pairs[k]->p2[X] - c2[X];
    v2[Y] = pairs[k]->p2[Y] - c2[Y];
    v2[Z] = pairs[k]->p2[Z] - c2[Z];

    cov[X][X] += v1[X] * v2[X];
    cov[X][Y] += v1[X] * v2[Y];
    cov[X][Z] += v1[X] * v2[Z];

    cov[Y][X] += v1[Y] * v2[X];
    cov[Y][Y] += v1[Y] * v2[Y];
    cov[Y][Z] += v1[Y] * v2[Z];

    cov[Z][X] += v1[Z] * v2[X];
    cov[Z][Y] += v1[Z] * v2[Y];
    cov[Z][Z] += v1[Z] * v2[Z];
  }

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      aij[i][j] = cov[i][j] - cov[j][i];
    }
  }

  tr = cov[X][X] + cov[Y][Y] + cov[Z][Z];

  m = m_get(4,4);
  q = m_get(4,4);
  v = v_get(4);

  m->me[0][0] = tr;

  m->me[1][0] = m->me[0][1] = aij[1][2];
  m->me[2][0] = m->me[0][2] = aij[2][0];
  m->me[3][0] = m->me[0][3] = aij[0][1];

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      m->me[i+1][j+1] = cov[i][j] + cov[j][i] - (i == j) * tr;

    }

  }

  
  v = symmeig( m, q, v );

  if (v->ve[0] > v->ve[1]) {

    if( v->ve[0] > v->ve[2]) {
      if(v->ve[0] > v->ve[3])
	i = 0;
      else
	i = 3;
    }
    else {
      if (v->ve[2] > v->ve[3])
	i = 2;
      else
	i =3;
    }
  }
  else {
    if (v->ve[1] > v->ve[2]) {
      if (v->ve[1] > v->ve[3])
	i = 1;
      else
	i = 3;
    }
    else {
      if (v->ve[2] > v->ve[3])
	i = 2;
      else
	i =3;
    }
  }

  quat[0] = q->me[0][i];
  quat[1] = q->me[1][i];
  quat[2] = q->me[2][i];
  quat[3] = q->me[3][i];

  convert_quat_to_mat (quat, rotmat);

  vapply (rotmat, c2, c2);
  rotmat[3][0] = c1[0] - c2[0];
  rotmat[3][1] = c1[1] - c2[1];
  rotmat[3][2] = c1[2] - c2[2];

  trans[X] = rotmat[3][0];
  trans[Y] = rotmat[3][1];
  trans[Z] = rotmat[3][2];

  dummy = m_free(m);
  dummy = m_free(q);
  dummy = v_free(v);
}



void convert_quat_to_mat(Quaternion q, Matrix mat) {

  float q00,q01,q02,q03;
  float q11,q12,q13;
  float q22,q23;
  float q33;
  
  q00 = q[0] * q[0];
  q01 = q[0] * q[1];
  q02 = q[0] * q[2];
  q03 = q[0] * q[3];
  
  q11 = q[1] * q[1];
  q12 = q[1] * q[2];
  q13 = q[1] * q[3];
  
  q22 = q[2] * q[2];
  q23 = q[2] * q[3];
  
  q33 = q[3] * q[3];
  
  mat[X][X] = q00 + q11 - q22 - q33;
  mat[X][Y] = 2 * (q12 - q03);
  mat[X][Z] = 2 * (q13 + q02);
  mat[X][W] = 0;
  
  mat[Y][X] =  2 * (q12 + q03);
  mat[Y][Y] =  q00 + q22 - q11 - q33;
  mat[Y][Z] = 2 * (q23 - q01);
  mat[Y][W] = 0;
  
  mat[Z][X] = 2 * (q13 - q02);
  mat[Z][Y] = 2 * (q23 + q01);
  mat[Z][Z] = q00 + q33 - q11 - q22;
  mat[Z][W] = 0;
  
  mat[W][X] = 0;
  mat[W][Y] = 0;
  mat[W][Z] = 0;
  mat[W][W] = 1;
}


